#!/usr/bin/python
#
# Copyright 2015 Canonical Ltd.
#
# Written by:
#   - Paul Collins <paul.collins@canonical.com>
#

import argparse
import logging
import os.path
import sys
import subprocess
import time
import types
import yaml

try:
    from shlex import quote as quote_for_shell
except ImportError:
    from pipes import quote as quote_for_shell

from .deploymentstatemanager import (
    DeploymentAlreadyAttemptedError,
    DeploymentStateManager,
)

from bzrlib.branch import Branch


class DoNotSeekTheTreasureError(Exception):
    """Thrown when the deployment cannot yet be attempted."""
    # XXX rename this, obviously
    pass


class Deployment(object):
    def __init__(self):
        # things
        pass

    @classmethod
    def from_config(cls, config, environment):
        deployment = cls()

        deployment.can_attempt_deployment_command = config.get('can-attempt-deployment-command')
        deployment.deployment_command = config['deployment-command']
        deployment.environment = environment.copy()

        deployment.spec_branch_location = spec_branch_location_for_environment(environment)
        deployment.cwd = None
        if config.get('deploy-from-spec-branch-directory'):
            deployment.cwd = deployment.spec_branch_location

        return deployment


    def can_be_attempted(self):
        if self.can_attempt_deployment_command:
            use_shell = (type(self.can_attempt_deployment_command)
                         is not types.ListType)
            try:
                subprocess.check_output(
                    self.can_attempt_deployment_command,
                    cwd=self.cwd,
                    env=self.environment,
                    shell=use_shell,
                )
            except subprocess.CalledProcessError as cpe:
                raise DoNotSeekTheTreasureError(
                    cpe.returncode, cpe.output)

        return True

    def deploy(self, deployment_log):
        # XXX enforce can_attempt() or just do it?
        use_shell = type(self.deployment_command) is not types.ListType
        subprocess.check_call(
            self.deployment_command,
            cwd=self.cwd,
            env=self.environment,
            shell=use_shell,
            stdout=deployment_log,
            stderr=subprocess.STDOUT,
        )


def envrc_as_dict(envrc_name):
    """Source the specified envrc in a subshell and return the resulting
       process environment as a dict."""

    env = subprocess.check_output(
        '. $HOME/.prodstack/envrc/{} >/dev/null 2>&1 && '
        '/usr/bin/env -0'.format(
            quote_for_shell(envrc_name)),
        shell=True,
    )

    return dict(e.split('=', 1) for e in env.split('\x00')[:-1])


def spec_branch_location_for_environment(environment):
    """Return $MOJO_SPEC if it is in the environment, or else run
       'mojo workspace-list' with the environment provided and return
       the spec branch location for $MOJO_WORKSPACE, or else return
       None if there is no match."""

    if 'MOJO_SPEC' in environment:
        return environment['MOJO_SPEC']
 
    mojo_output = subprocess.check_output(
        ['/usr/bin/mojo', 'workspace-list'],
        env=environment,
    )

    for line in mojo_output.split('\n'):
        try:
            workspace, spec_branch_location = line.split(': ', 1)
            if workspace == environment['MOJO_WORKSPACE']:
                return spec_branch_location
        except ValueError:
            pass

    return None


def get_last_revision(branch_location):
    branch = Branch.open(branch_location)
    return branch.last_revision()


def get_last_revision_info(branch_location):
    branch = Branch.open(branch_location)
    return branch.last_revision_info()


def do_pull(branch_location):
    target = Branch.open(branch_location)
    source = Branch.open(target.get_parent())
    result = target.pull(source)
    return result.old_revid != result.new_revid


def branch_is_deployable(spec_branch_location, deploy_tag):
    """Parse the top commit in the given branch and check if the
    commit message starts with the deploy_tag, in brackets."""

    branch = Branch.open(spec_branch_location)
    top_commit_summary = branch.repository.get_revisions(
        [branch.last_revision()])[0].get_summary()

    if top_commit_summary.startswith('[{}]'.format(deploy_tag)):
        return True

    return False


def parse_args():
    parser = argparse.ArgumentParser(
        description='Deploy from blessed branches.'
    )

    parser.add_argument(
        '--config-file',
        required=True,
    )
    parser.add_argument(
        '--verbose',
        default=False,
        action='store_true',
    )

    return parser.parse_args()


def main():
    args = parse_args()

    logging_level = args.verbose and logging.DEBUG or logging.INFO

    handler = logging.StreamHandler()
    handler.setLevel(logging_level)
    handler.setFormatter(
        logging.Formatter(
            fmt='%(asctime)s [%(levelname)s] %(message)s',
            datefmt='%Y-%m-%d %H:%M:%S',
        ))
    logger = logging.getLogger('deploy-from-blessed-branch')
    logger.setLevel(logging_level)
    logger.addHandler(handler)

    configuration = yaml.safe_load(open(args.config_file).read())
    dbfile_template = os.path.join(
        configuration['locations']['state-directory'],
        '{}.db'
    )

    for name, config in configuration['deployments'].items():
        logger.debug("processing {}".format(name))

        dsm = DeploymentStateManager(dbfile_template.format(name))

        logger.debug("loading envrc {}".format(config['environment-run-control']))
        environment = envrc_as_dict(config['environment-run-control'])

        spec_branch_location = spec_branch_location_for_environment(environment)
        logger.debug("spec branch location for {} = {}".format(
            config['environment-run-control'],
            spec_branch_location,
        ))

        with dsm.update_branch_from(get_last_revision(spec_branch_location)):
            logger.debug("updating {}".format(spec_branch_location))
            if do_pull(spec_branch_location):
                logger.info("changes pulled to {}".format(spec_branch_location))
            else:
                logger.debug("no changes to {}".format(spec_branch_location))

        logger.debug("checking if {} is deployable".format(spec_branch_location))

        if branch_is_deployable(spec_branch_location, config['deployable-revision-tag']):
            attempted_revno, attempted_revid = get_last_revision_info(spec_branch_location)
            try:
                with dsm.attempt_deployment(attempted_revid):
                    deployment_log_directory = config.get(
                        'log-directory',
                        configuration['locations']['log-directory']
                    )
                    deployment_log_file_name = os.path.join(
                        deployment_log_directory,
                        'deploy-{}-{}-r{}-{}.log'.format(
                            time.strftime('%Y%m%d-%H%M'),
                            name,
                            attempted_revno,
                            attempted_revid,
                        ),
                    )

                    deployment = Deployment.from_config(config, environment)

                    format_data = {
                        'name': name,
                        'attempted_revno': attempted_revno,
                        'attempted_revid': attempted_revid,
                        'spec_branch_location': deployment.spec_branch_location,
                        'deployment_command': deployment.deployment_command,
                        'can_attempt_deployment_command': deployment.can_attempt_deployment_command,
                        'deployment_log_file_name': deployment_log_file_name,
                    }

                    deployment.can_be_attempted()
                    deployment_log = open(deployment_log_file_name, 'a', 0)

                    logger.info(
                        "deploying {name} from {spec_branch_location} "
                        "with '{deployment_command}'; logging to "
                        "{deployment_log_file_name}".format(**format_data))

                    deployment_log.write(
                        "Begin deployment of {name} from {spec_branch_location} "
                        "r{attempted_revno} (revision {attempted_revid}) "
                        "with '{deployment_command}'\n\n".format(**format_data))

                    deployment.deploy(deployment_log)

                    logger.info(
                        "completed deploying {name} "
                        "from {spec_branch_location}".format(**format_data))

                    deployment_log.write(
                        "End deployment of {name} "
                        "from {spec_branch_location}\n".format(**format_data))

            except DoNotSeekTheTreasureError as dnstte:
                logger.debug(
                    "{name} r{attempted_revno} (revision {attempted_revid} "
                    "is not ready to be deployed yet".format(**format_data))
                logger.debug(
                    "'{can_attempt_deployment_command}' exited "
                    "with status {returncode}".format(
                        returncode=dnstte.args[0],
                        **format_data))
                if dnstte.args[1]:
                    logger.debug("with the following output:")
                    logger.debug(
                        ('\n' + dnstte.args[1])
                        .rstrip('\n')
                        .replace('\n','\n> ')
                    )
                dsm.unattempt_deploy(attempted_revid)

            except DeploymentAlreadyAttemptedError:
                verbed = 'attempted or in progress'
                if dsm.completed(attempted_revid):
                    verbed = 'completed'
                logger.debug('deployment of {} from {} r{} (revision {}) already {}; nothing to do'.format(
                    name,
                    spec_branch_location,
                    attempted_revno,
                    attempted_revid,
                    verbed,
                ))

            except Exception:
                logger.error('deployment failed; deployment log follows:')
                logger.error(
                    ('\n' + open(deployment_log_file_name).read())
                    .rstrip('\n')
                    .replace('\n','\n> ')
                )
                logger.error('deployment log ends')
                logger.error('exception caught during deployment; re-raising')
                raise
        else:
            logger.debug("skipping {} as the top commit is not tagged for deployment".format(name))
