#
# Copyright 2015 Canonical Ltd.
#
# Written by:
#  - Paul Collins <paul.collins@canonical.com>
#

# This module is used to safely coordinate activities via a sqlite
# database file with context managers.  Currently it supports:
#
#  - updating a bzr branch, exclusion based on the starting revid
#
#    Since the branch can be updated at any time from a given revid,
#    we remove the revid from the table afterwards so that we always
#    attempt to update branches.
#
#  - deploying from a bzr branch, exclusion based on the final revid
#
#    In general we want to deploy a given revid only once.  The
#    context manager does not remove the revid from the
#    attempted_deployments table even when an error occurs during the
#    deployment.  We can add more smarts here if desired to account
#    from transient deployment failures (although these should be rare).
#    When deployments succeed, we add the revid to completed_deployments.

import sqlite3


UPDATING_BRANCH_FROM_DDL = """
CREATE TABLE IF NOT EXISTS update_branch_from (ubf_revid TEXT PRIMARY KEY UNIQUE)
"""

ATTEMPTED_DEPLOYMENTS_DDL = """
CREATE TABLE IF NOT EXISTS attempted_deployments (ad_revid TEXT PRIMARY KEY UNIQUE)
"""

COMPLETED_DEPLOYMENTS_DDL = """
CREATE TABLE IF NOT EXISTS completed_deployments (cd_revid TEXT PRIMARY KEY UNIQUE)
"""


class BranchIsAlreadyUpdatingError(Exception):
    pass


class DeploymentAlreadyAttemptedError(Exception):
    pass


class AttemptedDeploymentContextManager(object):
    def __init__(self, dsm, revid):
        self._dsm = dsm
        self._revid = revid

    def __enter__(self):
        # XXX sanity check revid here?
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        if (exc_type, exc_val, exc_tb) != (None, None, None):
            # We let exceptions bubble up freely.
            return False

        self._dsm._did_complete_deployment(self._revid)
        return True


class BranchUpdatingContextManager(object):
    def __init__(self, dsm, revid):
        self._dsm = dsm
        self._revid = revid

    def __enter__(self):
        pass

    def __exit__(self, exc_type, exc_val, exc_tb):
        self._dsm._did_update_branch_from(self._revid)

        return False


class DeploymentStateManager(object):
    def __init__(self, dbfile):
        self._db = sqlite3.connect(dbfile)
        cur = self._db.cursor()
        cur.execute(UPDATING_BRANCH_FROM_DDL)
        cur.execute(ATTEMPTED_DEPLOYMENTS_DDL)
        cur.execute(COMPLETED_DEPLOYMENTS_DDL)
        self._db.commit()

    def update_branch_from(self, revid):
        try:
            cur = self._db.cursor()
            cur.execute('INSERT INTO update_branch_from (ubf_revid) values (?)', [revid])
            self._db.commit()
        except sqlite3.IntegrityError as ie:
            self._db.rollback()
            raise BranchIsAlreadyUpdatingError(revid)

        return BranchUpdatingContextManager(self, revid)

    def _did_update_branch_from(self, revid):
        cur = self._db.cursor()
        cur.execute('DELETE FROM update_branch_from where ubf_revid = ?', [revid])
        self._db.commit()

    def attempt_deployment(self, revid):
        try:
            cur = self._db.cursor()
            cur.execute('INSERT INTO attempted_deployments (ad_revid) values (?)', [revid])
            self._db.commit()
        except sqlite3.IntegrityError as ie:
            self._db.rollback()
            raise DeploymentAlreadyAttemptedError(revid)

        return AttemptedDeploymentContextManager(self, revid)

    def unattempt_deploy(self, revid):
        cur = self._db.cursor()
        cur.execute('DELETE FROM attempted_deployments where ad_revid = ?', [revid])
        self._db.commit()

    def _did_complete_deployment(self, revid):
        try:
            # XXX sanity check vs. attempted_deployments?
            cur = self._db.cursor()
            cur.execute('INSERT INTO completed_deployments (cd_revid) values (?)', [revid])
            self._db.commit()
        except sqlite3.IntegrityError as ie:
            # XXX failure here is super problematic, since it must
            # mean we raced another instance somehow, which the
            # attempted_deployments table is supposed to prevent.
            self._db.rollback()
            raise

    def completed(self, revid):
        return len(self._db.execute(
            'SELECT cd_revid FROM completed_deployments '
            'WHERE cd_revid = ?', [revid]).fetchall()) == 1
